import 'connection_pool.dart';
import 'table_meta.dart';
import 'xdatabase_base.dart';

class NotSupportedException extends Error {}

class DeclarationException implements Exception {
  String message;
  DeclarationException(this.message);
  @override
  String toString() {
    return message;
  }
}

typedef HandleObjectMap = Map<String, dynamic> Function(dynamic obj);
typedef HandleMapObject = dynamic Function(Map<String, dynamic> row);

class Query<T> {
  QueryAble ctx;
  String statement;
  Orm<T> ormap;
  late TableDeclaration tableInfo;
  Query(this.statement, this.ctx, this.ormap) {
    tableInfo = DeclarationHolder().byTable(T.toString())!;
  }

  String trimSql() {
    final idx = statement.toLowerCase().indexOf('where');
    return statement.substring(idx);
  }

  Future<int> update(Bean<T> obj) async {
    final sql = ormap.updateSql(trimSql());
    final params = obj.asMap();
    return await ctx.execute(sql, substitutionValues: params);
  }

  Future<int> delete(Bean<T> obj) async {
    final sql = ormap.deleteSql(trimSql());
    final params = obj.asMap();
    return await ctx.execute(sql, substitutionValues: params);
  }

  Future<List<T>> fetch(Bean<T> obj, {bool all = false}) async {
    List<Map<String, Map<String, dynamic>>> resultSet;
    if (!all) {
      final params = obj.asMap();
      resultSet =
          await ctx.mappedResultsQuery(statement, substitutionValues: params);
    } else {
      resultSet = await ctx.mappedResultsQuery(statement);
    }
    if (resultSet.isNotEmpty) {
      return resultSet
          .map((row) => obj.fromMap(row[tableInfo.tableName]!))
          .toList();
    } else {
      return <T>[];
    }
  }
}

class Orm<T> {
  late TableDeclaration tableInfo;
  QueryAble ctx;
  Orm(this.ctx) {
    final _tableInfo = DeclarationHolder().byTable(T.toString());
    if (_tableInfo == null) {
      throw DeclarationException('Table $T not register!');
    }
    tableInfo = _tableInfo;
  }

  String _insertSQL() {
    var lines = [
      'INSERT INTO ${tableInfo.tableName}("',
      tableInfo.fields.map((fd) => fd.fieldName).toList().join('","'),
      '") VALUES (',
      tableInfo.fields.map((fd) => '@${fd.fieldName}').toList().join(','),
      ');'
    ];
    return lines.join('');
  }

  Future<int> insert(Bean<T> obj) async {
    final sql = _insertSQL();
    final param = obj.asMap();
    print(param);
    return await ctx.execute(sql, substitutionValues: param);
  }

  Query<T> query(String statement) {
    return Query(
        statement.replaceAll('[table]', tableInfo.tableName), ctx, this);
  }

  String statementByPk() {
    final pks = tableInfo.fields.where((element) => element.primary).toList();
    if (pks.isEmpty) {
      throw NotSupportedException();
    }
    return 'SELECT * FROM "${tableInfo.tableName}" WHERE "${pks.first.fieldName}"=@${pks.first.fieldName};';
  }

  String deleteSql(String statement) {
    return 'DELETE FROM "${tableInfo.tableName}" $statement';
  }

  String updateSql(String statement) {
    final lines = <String>[
      'UPDATE "${tableInfo.tableName}" SET ',
      tableInfo.fields
          .map((FieldDeclaration fd) => '"${fd.fieldName}"=@${fd.fieldName}')
          .toList()
          .join(', '),
      ' $statement;'
    ];
    return lines.join('');
  }

  Future<int> deleteByPk(Bean<T> obj) {
    final sql = deleteSql(statementByPk());
    final params = obj.asMap();
    return ctx.execute(sql, substitutionValues: params);
  }

  Future<int> updateByPk(Bean<T> obj) {
    final sql = updateSql(statementByPk());
    final params = obj.asMap();
    return ctx.execute(sql, substitutionValues: params);
  }

  Future<T?> byPrimary(Bean<T> obj) async {
    final sql = statementByPk();
    final queryObj = query(sql);
    final results = await queryObj.fetch(obj);
    if (results.isNotEmpty) {
      return results.first;
    }
  }
}
